# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

log = logging.getLogger(__name__)


class BaseProcessor(object):

    dimensions = {'service': 'monitoring', 'component': 'monasca-notification'}

    @staticmethod
    def _add_to_queue(queue, queue_name, msg):
        """Warns on full queue then does a blocking push to the queue.
        """
        if queue.full():
            log.warn('Queue %s is full, publishing is blocked' % queue_name)
        queue.put(msg)
        log.debug("Put message %s on queue %s" % (msg, queue_name))



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the AlarmProcessor"""

import collections
import json
import mock
import multiprocessing
import Queue
import time
import unittest

from monasca_notification.notification import Notification
from monasca_notification.processors import alarm_processor

alarm_tuple = collections.namedtuple('alarm_tuple', ['offset', 'message'])
message_tuple = collections.namedtuple('message_tuple', ['value'])


class TestAlarmProcessor(unittest.TestCase):
    def setUp(self):
        self.alarm_queue = multiprocessing.Queue(10)
        self.notification_queue = multiprocessing.Queue(10)
        self.finished_queue = multiprocessing.Queue(10)
        self.log_queue = multiprocessing.Queue(10)

    def _create_raw_alarm(self, partition, offset, message):
        """Create a raw alarm, with the given message dictionary.
        """
        json_msg = json.dumps({'alarm-transitioned': message})
        msg_tuple = message_tuple(json_msg)
        return [partition, alarm_tuple(offset, msg_tuple)]

    @mock.patch('MySQLdb.connect')
    @mock.patch('monasca_notification.processors.alarm_processor.log')
    def _run_alarm_processor(self, queue, sql_response, mock_log, mock_mysql):
        """Runs a mocked alarm processor reading from queue while running, returns (queue_message, log_message)
        """
        # Since the log runs in another thread I can mock it directly, instead change the methods to put to a queue
        mock_log.warn = self.log_queue.put
        mock_log.error = self.log_queue.put

        # Setup the sql response
        if sql_response is not None:
            mock_mysql.return_value = mock_mysql
            mock_mysql.cursor.return_value = mock_mysql
            mock_mysql.__iter__.return_value = sql_response

        processor = alarm_processor.AlarmProcessor(self.alarm_queue, self.notification_queue,
                                                   self.finished_queue, 600, 'mysql_host', 'mysql_user',
                                                   'mysql_passwd', 'dbname')

        p_thread = multiprocessing.Process(target=processor.run)
        p_thread.start()
        try:
            queue_msg = queue.get(timeout=2)
        except Queue.Empty:
            queue_msg = None
        p_thread.terminate()

        try:
            log_msg = self.log_queue.get(timeout=1)
        except Queue.Empty:
            log_msg = None
        return queue_msg, log_msg

    def test_invalid_alarm(self):
        """Invalid Alarms, should log and error and push to the finished queue."""
        self.alarm_queue.put(self._create_raw_alarm(0, 1, {'invalid': 'invalid_alarm'}))
        finished, log_msg = self._run_alarm_processor(self.finished_queue, None)

        self.assertTrue(finished == (0, 1))
        self.assertTrue(log_msg.startswith('Invalid Alarm format'))

    def test_old_timestamp(self):
        """Should cause the alarm_ttl to fire log a warning and push to finished queue."""
        alarm_dict = {"tenantId": "0", "alarmDefinitionId": "0", "alarmId": "1", "alarmName": "test Alarm",
                      "oldState": "OK", "newState": "ALARM", "stateChangeReason": "I am alarming!",
                      "timestamp": 1375346830, "actionsEnabled": 1, "metrics": "cpu_util"}
        self.alarm_queue.put(self._create_raw_alarm(0, 2, alarm_dict))
        finished, log_msg = self._run_alarm_processor(self.finished_queue, None)

        self.assertTrue(finished == (0, 2))
        self.assertTrue(log_msg.startswith('Received alarm older than the ttl'))

    def test_no_notifications(self):
        """Test an alarm with no defined notifications
        """
        alarm_dict = {"tenantId": "0", "alarmDefinitionId": "0", "alarmId": "1", "alarmName": "test Alarm",
                      "oldState": "OK", "newState": "ALARM", "stateChangeReason": "I am alarming!",
                      "timestamp": time.time(), "actionsEnabled": 1, "metrics": "cpu_util"}
        self.alarm_queue.put(self._create_raw_alarm(0, 3, alarm_dict))
        finished, log_msg = self._run_alarm_processor(self.finished_queue, None)

        self.assertTrue(finished == (0, 3))

    def test_valid_notification(self):
        """Test a valid notification, being put onto the notification_queue
        """
        alarm_dict = {"tenantId": "0", "alarmDefinitionId": "0", "alarmId": "1", "alarmName": "test Alarm",
                      "oldState": "OK", "newState": "ALARM", "stateChangeReason": "I am alarming!",
                      "timestamp": time.time(), "actionsEnabled": 1, "metrics": "cpu_util"}
        self.alarm_queue.put(self._create_raw_alarm(0, 4, alarm_dict))
        sql_response = [['test notification', 'EMAIL', 'me@here.com']]
        finished, log_msg = self._run_alarm_processor(self.notification_queue, sql_response)

        test_notification = Notification('email', 0, 4, 'test notification', 'me@here.com', alarm_dict)

        self.assertTrue(finished == [test_notification])

    def test_two_valid_notifications(self):
        alarm_dict = {"tenantId": "0", "alarmDefinitionId": "0", "alarmId": "1", "alarmName": "test Alarm",
                      "oldState": "OK", "newState": "ALARM", "stateChangeReason": "I am alarming!",
                      "timestamp": time.time(), "actionsEnabled": 1, "metrics": "cpu_util"}
        self.alarm_queue.put(self._create_raw_alarm(0, 5, alarm_dict))
        sql_response = [['test notification', 'EMAIL', 'me@here.com'], ['test notification2', 'EMAIL', 'me@here.com']]
        finished, log_msg = self._run_alarm_processor(self.notification_queue, sql_response)

        test_notification = Notification('email', 0, 5, 'test notification', 'me@here.com', alarm_dict)
        test_notification2 = Notification('email', 0, 5, 'test notification2', 'me@here.com', alarm_dict)

        self.assertTrue(finished == [test_notification, test_notification2])



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the StateTracker"""

import kafka.common
import mock
import multiprocessing
import threading
import time
import unittest

from monasca_notification import state_tracker


class TestStateTracker(unittest.TestCase):
    def setUp(self):
        self.finished_queue = multiprocessing.Queue(10)
        with mock.patch('kazoo.client.KazooClient') as self.mock_zk:
            self.mock_zk.return_value = self.mock_zk
            with mock.patch('kafka.client.KafkaClient') as self.mock_kafka:
                self.tracker = state_tracker.KafkaStateTracker(self.finished_queue, 'kafka_url', 'group',
                                                               'topic', 1, 'zookeeper_url')
                self.mock_kafka.return_value = self.mock_kafka

        self.tracker.has_lock = True
        self.tracker_thread = threading.Thread(target=self.tracker.run)
        self.tracker_thread.daemon = True  # needed for the thread to properly exit

    def _feed_queue(self, test_list):
        """Feed the queue and start the state tracker.
        """
        for item in test_list:
            self.finished_queue.put(item)

        self.tracker._offsets = {item[0]: 0 for item in test_list}
        self.tracker_thread.start()
        time.sleep(1)

    def test_ordered(self):
        """Test a series of ordered finished offsets to make sure they are updated correctly
        """
        test_list = [(0, 1), (0, 2), (1, 1), (1, 2)]
        self._feed_queue(test_list)

        expected_calls = [mock.call().send_offset_commit_request('group',
                                                                 [kafka.common.OffsetCommitRequest('topic', partition,
                                                                                                   value, None)])
                          for partition, value in test_list]
        set_calls = [call for call in self.mock_kafka.mock_calls
                     if call.__str__().startswith("call().send_offset_commit_request('")]

        self.assertTrue(expected_calls == set_calls)

    def test_unordered(self):
        """Test a series of unordered finished offsets to make sure they are updated in order
        """
        unordered_test_list = [(0, 2), (0, 1), (1, 2), (1, 1)]
        self._feed_queue(unordered_test_list)

        commit_list = [(0, 2), (1, 2)]
        expected_calls = [mock.call().send_offset_commit_request('group',
                                                                 [kafka.common.OffsetCommitRequest('topic', partition,
                                                                                                   value, None)])
                          for partition, value in commit_list]
        set_calls = [call for call in self.mock_kafka.mock_calls
                     if call.__str__().startswith("call().send_offset_commit_request('")]

        self.assertTrue(expected_calls == set_calls)

    def test_unordered_with_lag(self):
        """Test a series of unordered finished offsets with lag being received
        """
        unordered_quick = [(0, 2), (1, 2)]
        self._feed_queue(unordered_quick)

        time.sleep(1)  # sleep the lag period. and feed more
        unordered_slow = [(0, 3), (1, 3)]
        for item in unordered_slow:
            self.finished_queue.put(item)

        time.sleep(1)  # wait for processing

        commit_list = [(0, 3), (1, 3)]
        expected_calls = [mock.call().send_offset_commit_request('group',
                                                                 [kafka.common.OffsetCommitRequest('topic', partition,
                                                                                                   value, None)])
                          for partition, value in commit_list]
        set_calls = [call for call in self.mock_kafka.mock_calls
                     if call.__str__().startswith("call().send_offset_commit_request('")]

        self.assertTrue(expected_calls == set_calls)



#!/usr/bin/env python
# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Notification Engine
    This engine reads alarms from Kafka and then notifies the customer using their configured notification method.
"""

import logging
import logging.config
import multiprocessing
import os
import signal
from state_tracker import KafkaStateTracker
import sys
import threading
import time
import yaml

from processors.alarm_processor import AlarmProcessor
from processors.kafka_consumer import KafkaConsumer
from processors.notification_processor import NotificationProcessor
from processors.sent_notification_processor import SentNotificationProcessor

log = logging.getLogger(__name__)
processors = []  # global list to facilitate clean signal handling
exiting = False


def clean_exit(signum, frame=None):
    """Exit all processes attempting to finish uncommited active work before exit.
         Can be called on an os signal or no zookeeper losing connection.
    """
    global exiting
    if exiting:
        # Since this is set up as a handler for SIGCHLD when this kills one child it gets another signal, the global
        # exiting avoids this running multiple times.
        log.debug('Exit in progress clean_exit received additional signal %s' % signum)
        return

    log.info('Received signal %s, beginning graceful shutdown.' % signum)
    exiting = True

    # the final processor is the sent_notification processor, skip it and the tracker both of which should
    # finish processing of any already sent notifications.
    for process in processors[:-1]:
        try:
            if process.is_alive():
                process.terminate()  # Sends sigterm which any processes after a notification is sent attempt to handle
        except Exception:
            pass

    tracker.stop = True
    max_wait_count = 5
    while tracker.has_lock:
        if max_wait_count == 0:
            log.debug('Max wait reached, proceeding to kill processes')
            break
        log.debug('Waiting for all active processing to stop.')
        max_wait_count -= 1
        time.sleep(2)

    # Kill everything, that didn't already die
    for child in multiprocessing.active_children():
        log.debug('Killing pid %s' % child.pid)
        try:
            os.kill(child.pid, signal.SIGKILL)
        except Exception:
            pass

    sys.exit(0)


def main(argv=None):
    if argv is None:
        argv = sys.argv
    if len(argv) == 2:
        config_file = argv[1]
    elif len(argv) > 2:
        print("Usage: " + argv[0] + " <config_file>")
        print("Config file defaults to /etc/monasca/notification.yaml")
        return 1
    else:
        config_file = '/etc/monasca/notification.yaml'

    config = yaml.load(open(config_file, 'r'))

    # Setup logging
    logging.config.dictConfig(config['logging'])

    # Create the queues
    alarms = multiprocessing.Queue(config['queues']['alarms_size'])
    notifications = multiprocessing.Queue(config['queues']['notifications_size'])  # [notification_object, ]
    sent_notifications = multiprocessing.Queue(config['queues']['sent_notifications_size'])  # [notification_object, ]
    finished = multiprocessing.Queue(config['queues']['finished_size'])  # Data is of the form (partition, offset)

    # State Tracker - Used for tracking the progress of fully processed alarms and the zookeeper lock
    global tracker  # Set to global for use in the cleanup function
    tracker = KafkaStateTracker(finished, config['kafka']['url'], config['kafka']['group'],
                                config['kafka']['alarm_topic'], config['kafka']['max_offset_lag'],
                                config['zookeeper']['url'])
    tracker.lock(clean_exit)  # Only begin if we have the processing lock
    tracker_thread = threading.Thread(target=tracker.run)

    # Define processors
    # KafkaConsumer
    kafka = multiprocessing.Process(
        target=KafkaConsumer(
            alarms,
            config['kafka']['url'],
            config['kafka']['group'],
            config['kafka']['alarm_topic']
        ).run
    )
    processors.append(kafka)

    # AlarmProcessors
    alarm_processors = []
    if 'ssl' in config['mysql']:
        ssl_config = config['mysql']['ssl']
    else:
        ssl_config = None
    for i in range(config['processors']['alarm']['number']):
        alarm_processors.append(multiprocessing.Process(
            target=AlarmProcessor(
                alarms,
                notifications,
                finished,
                config['processors']['alarm']['ttl'],
                config['mysql']['host'],
                config['mysql']['user'],
                config['mysql']['passwd'],
                config['mysql']['db'],
                ssl_config
            ).run),
        )
    processors.extend(alarm_processors)

    # NotificationProcessors
    notification_processors = []
    for i in range(config['processors']['notification']['number']):
        notification_processors.append(multiprocessing.Process(
            target=NotificationProcessor(
                notifications,
                sent_notifications,
                finished,
                config['email'],
                config['webhook'],
                config['pagerduty']
            ).run),
        )
    processors.extend(notification_processors)

    # SentNotificationProcessor
    sent_notification_processor = multiprocessing.Process(
        target=SentNotificationProcessor(
            sent_notifications,
            finished,
            config['kafka']['url'],
            config['kafka']['notification_topic']
        ).run
    )
    processors.append(sent_notification_processor)

    # Start
    try:
        log.info('Starting processes')
        for process in processors:
            process.start()

        # The offset tracker runs in a thread so the signal handler can run concurrently and cleanly shutdown
        tracker_thread.start()

        # The signal handlers must be added after the processes start otherwise they run on all processes
        signal.signal(signal.SIGCHLD, clean_exit)
        signal.signal(signal.SIGINT, clean_exit)
        signal.signal(signal.SIGTERM, clean_exit)

        # If the tracker fails exit
        while True:
            if tracker_thread.is_alive():
                time.sleep(3)
            else:
                tracker.has_lock = False
                clean_exit('tracker died', None)
    except Exception:
        log.exception('Error! Exiting.')
        for process in processors:
            process.terminate()

if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import monascastatsd
import MySQLdb
import time

from monasca_notification.notification import Notification
from monasca_notification.notification_exceptions import AlarmFormatError
from monasca_notification.processors.base import BaseProcessor


log = logging.getLogger(__name__)


class AlarmProcessor(BaseProcessor):
    def __init__(
            self, alarm_queue, notification_queue, finished_queue,
            alarm_ttl, mysql_host, mysql_user, mysql_passwd, dbname, mysql_ssl=None):
        self.alarm_queue = alarm_queue
        self.alarm_ttl = alarm_ttl
        self.notification_queue = notification_queue
        self.finished_queue = finished_queue
        self.statsd = monascastatsd.Client(name='monasca',
                                           dimensions=BaseProcessor.dimensions)
        try:
            self.mysql = MySQLdb.connect(host=mysql_host, user=mysql_user,
                                         passwd=unicode(mysql_passwd).encode('utf-8'), db=dbname, ssl=mysql_ssl)
            self.mysql.autocommit(True)
        except:
            log.exception('MySQL connect failed')
            raise

    @staticmethod
    def _parse_alarm(alarm_data):
        """Parse the alarm message making sure it matches the expected format.
        """
        expected_fields = [
            'actionsEnabled',
            'alarmId',
            'alarmDefinitionId',
            'alarmName',
            'newState',
            'oldState',
            'stateChangeReason',
            'tenantId',
            'timestamp'
        ]

        json_alarm = json.loads(alarm_data)
        alarm = json_alarm['alarm-transitioned']
        for field in expected_fields:
            if field not in alarm:
                raise AlarmFormatError('Alarm data missing field %s' % field)
        if ('tenantId' not in alarm) or ('alarmId' not in alarm):
            raise AlarmFormatError

        return alarm

    def _alarm_is_valid(self, alarm):
        """Check if the alarm is enabled and is within the ttl, return True in that case
        """
        if not alarm['actionsEnabled']:
            log.debug('Actions are disabled for this alarm.')
            return False

        alarm_age = time.time() - alarm['timestamp']  # Should all be in seconds since epoch
        if (self.alarm_ttl is not None) and (alarm_age > self.alarm_ttl):
            log.warn('Received alarm older than the ttl, skipping. Alarm from %s' % time.ctime(alarm['timestamp']))
            return False

        return True

    def run(self):
        """Check the notification setting for this project in mysql then create the appropriate notification or
             add to the finished_queue
        """
        failed_parse_count = self.statsd.get_counter(name='alarms_failed_parse_count')
        no_notification_count = self.statsd.get_counter(name='alarms_no_notification_count')
        notification_count = self.statsd.get_counter(name='created_count')
        db_time = self.statsd.get_timer()

        cur = self.mysql.cursor()

        while True:
            raw_alarm = self.alarm_queue.get()
            partition = raw_alarm[0]
            offset = raw_alarm[1].offset
            try:
                alarm = self._parse_alarm(raw_alarm[1].message.value)
            except Exception as e:  # This is general because of a lack of json exception base class
                failed_parse_count += 1
                log.error("Invalid Alarm format skipping partition %d, offset %d\nErrror%s" % (partition, offset, e))
                self._add_to_queue(self.finished_queue, 'finished', (partition, offset))
                continue

            log.debug("Read alarm from alarms sent_queue. Partition %d, Offset %d, alarm data %s"
                      % (partition, offset, alarm))

            if not self._alarm_is_valid(alarm):
                no_notification_count += 1
                self._add_to_queue(self.finished_queue, 'finished', (partition, offset))
                continue

            try:
                with db_time.time('config_db_time'):
                    cur.execute("""SELECT name, type, address
                                   FROM alarm_action as aa
                                   JOIN notification_method as nm ON aa.action_id = nm.id
                                   WHERE aa.alarm_definition_id = %s and aa.alarm_state = %s""",
                                [alarm['alarmDefinitionId'], alarm['newState']])
            except MySQLdb.Error:
                log.exception('Mysql Error')
                raise

            notifications = [
                Notification(row[1].lower(), partition, offset, row[0], row[2], alarm) for row in cur]

            if len(notifications) == 0:
                no_notification_count += 1
                log.debug('No notifications found for this alarm, partition %d, offset %d, alarm data %s'
                          % (partition, offset, alarm))
                self._add_to_queue(self.finished_queue, 'finished', (partition, offset))
            else:
                notification_count += len(notifications)
                self._add_to_queue(self.notification_queue, 'notifications', notifications)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import email.mime.text
import json
import logging
import monascastatsd
import requests
import smtplib
import sys
import time

from monasca_notification.processors.base import BaseProcessor

log = logging.getLogger(__name__)


class NotificationProcessor(BaseProcessor):

    def __init__(self, notification_queue,
                 sent_notification_queue, finished_queue,
                 email_config, webhook_config, pagerduty_config):
        self.notification_queue = notification_queue
        self.sent_notification_queue = sent_notification_queue
        self.finished_queue = finished_queue

        self.email_config = email_config

        self.webhook_config = {'timeout': 5}
        self.webhook_config.update(webhook_config)

        self.pagerduty_config = {
            'timeout': 5,
            'url': 'https://events.pagerduty.com/generic/2010-04-15/create_event.json'}
        self.pagerduty_config.update(pagerduty_config)

        self.smtp = None
        self._smtp_connect()

        # Types as key, method used to process that type as value
        self.notification_types = {'email': self._send_email,
                                   'webhook': self._post_webhook,
                                   'pagerduty': self._post_pagerduty}

        self.statsd = monascastatsd.Client(name='monasca', dimensions=BaseProcessor.dimensions)

    def _create_msg(self, hostname, notification):
        """Create two kind of messages:
        1. Notifications that include metrics with a hostname as a dimension. There may be more than one hostname.
           We will only report the hostname if there is only one.
        2. Notifications that do not include metrics and therefore no hostname. Example: API initiated changes.
           * A third notification type which include metrics but do not include a hostname will
           be treated as type #2.
        """
        if len(hostname) == 1:  # Type 1
            msg = email.mime.text.MIMEText("On host \"%s\" %s\n\nAlarm \"%s\" transitioned to the %s state at %s UTC"
                                           % (hostname[0],
                                               notification.message.lower(),
                                               notification.alarm_name,
                                               notification.state,
                                               time.asctime(time.gmtime(notification.alarm_timestamp))) +
                                           "\nalarm_id: %s" % notification.alarm_id)

            msg['Subject'] = "%s \"%s\" for Host: %s" % (notification.state, notification.alarm_name, hostname[0])

        else:  # Type 2
            msg = email.mime.text.MIMEText("%s\n\nAlarm \"%s\" transitioned to the %s state at %s UTC\nAlarm_id: %s"
                                           % (notification.message,
                                              notification.alarm_name,
                                              notification.state,
                                              time.asctime(time.gmtime(notification.alarm_timestamp)),
                                              notification.alarm_id))
            msg['Subject'] = "%s \"%s\" " % (notification.state, notification.alarm_name)

        msg['From'] = self.email_config['from_addr']
        msg['To'] = notification.address

        return msg

    def _send_email(self, notification):
        """Send the notification via email
             Returns the notification upon success, None upon failure
        """

        # Get the "hostname" from the notification metrics if there is one
        hostname = []

        for metric in notification.metrics:
            for dimension in metric['dimensions']:
                if 'hostname' in dimension:
                    if not metric['dimensions']['%s' % dimension] in hostname[:]:
                        hostname.append(metric['dimensions']['%s' % dimension])

        # Generate the message
        msg = self._create_msg(hostname, notification)

        # email the notification
        try:
            self.smtp.sendmail(self.email_config['from_addr'], notification.address, msg.as_string())
            log.debug('Sent email to %s, notification %s' % (notification.address, notification.to_json()))
        except smtplib.SMTPServerDisconnected:
            log.debug('SMTP server disconnected. Will reconnect and retry message.')
            self._smtp_connect()
            try:
                self.smtp.sendmail(self.email_config['from_addr'], notification.address, msg.as_string())
                log.debug('Sent email to %s, notification %s' % (notification.address, notification.to_json()))
            except smtplib.SMTPException as e:
                log.error("Error sending Email Notification:%s\nError:%s" % (notification.to_json(), e))
        except smtplib.SMTPException as e:
            log.error("Error sending Email Notification:%s\nError:%s" % (notification.to_json(), e))
        else:
            return notification

    def _smtp_connect(self):
        """Connect to the smtp server
        """
        log.info('Connecting to Email Server %s' % self.email_config['server'])
        smtp = smtplib.SMTP(
            self.email_config['server'], self.email_config['port'], timeout=self.email_config['timeout'])

        if self.email_config['user'] is not None:
            smtp.login(self.email_config['user'], self.email_config['password'])

        self.smtp = smtp

    def _post_webhook(self, notification):
        """Send the notification via webhook
            Posts on the given url
        """

        log.info(
            "Notifying alarm %(alarm_id)s to %(current)s with action %(action)s" %
            ({'alarm_id': notification.alarm_name,
              'current': notification.state,
              'action': notification.address}))
        body = {'alarm_id': notification.alarm_id}
        headers = {'content-type': 'application/json'}

        url = notification.address

        try:
            # Posting on the given URL
            result = requests.post(url=url,
                                   data=body,
                                   headers=headers,
                                   timeout=self.webhook_config['timeout'])
            if result.status_code in range(200, 300):
                log.info("Notification successfully posted.")
                return notification
            else:
                log.error("Received an HTTP code %s when trying to post on URL %s." % (result.status_code, url))
        except:
            log.error("Error trying to post on URL %s: %s." % (url, sys.exc_info()[0]))

    def _post_pagerduty(self, notification):
        """Send pagerduty notification
        """

        url = self.pagerduty_config['url']
        headers = {"content-type": "application/json"}
        body = {"service_key": notification.address,
                "event_type": "trigger",
                "description": notification.message,
                "client": "Monasca",
                "client_url": "",
                "details": {"alarm_id": notification.alarm_id,
                            "alarm_name": notification.alarm_name,
                            "current": notification.state,
                            "message": notification.message}}

        try:
            result = requests.post(url=url,
                                   data=json.dumps(body),
                                   headers=headers,
                                   timeout=self.pagerduty_config['timeout'])

            valid_http_codes = [200, 201, 204]
            if result.status_code in valid_http_codes:
                return notification

            log.error("Error with pagerduty request. key=<%s> response=%s"
                      % (notification.address, result.status_code))
        except:
            log.error("Exception on pagerduty request. key=<%s> exception=%s"
                      % (notification.address, sys.exc_info()[0]))

    def run(self):
        """Send the notifications
             For each notification in a message it is sent according to its type.
             If all notifications fail the alarm partition/offset are added to the the finished queue
        """
        counters = {'email': self.statsd.get_counter(name='sent_smtp_count'),
                    'webhook': self.statsd.get_counter(name='sent_webhook_count'),
                    'pagerduty': self.statsd.get_counter(name='sent_pagerduty_count')}
        timers = {'email': self.statsd.get_timer(),
                  'webhook': self.statsd.get_timer(),
                  'pagerduty': self.statsd.get_timer()}
        invalid_type_count = self.statsd.get_counter(name='invalid_type_count')
        sent_failed_count = self.statsd.get_counter(name='sent_failed_count')

        while True:
            notifications = self.notification_queue.get()
            sent_notifications = []
            for notification in notifications:
                if notification.type not in self.notification_types:
                    log.warn('Notification type %s is not a valid type' % notification.type)
                    invalid_type_count += 1
                    continue
                else:
                    timer_name = notification.type + '_time'
                    with timers[notification.type].time(timer_name):
                        sent = self.notification_types[notification.type](notification)

                    if sent is None:
                        sent_failed_count += 1
                    else:
                        sent.notification_timestamp = time.time()
                        sent_notifications.append(sent)
                        counters[notification.type] += 1
            if len(sent_notifications) == 0:  # All notifications failed
                self._add_to_queue(
                    self.finished_queue, 'finished', (notifications[0].src_partition, notifications[0].src_offset))
            else:
                self._add_to_queue(self.sent_notification_queue, 'sent_notification', sent_notifications)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json


class Notification(object):
    """An abstract base class used to define the notification interface
       and common functions
    """
    __slots__ = (
        'address',
        'alarm_id',
        'alarm_name',
        'alarm_timestamp',
        'message',
        'name',
        'notification_timestamp',
        'src_partition',
        'src_offset',
        'state',
        'tenant_id',
        'type',
        'metrics'
    )

    def __init__(self, ntype, src_partition, src_offset, name, address, alarm):
        """Setup the notification object
             The src_partition and src_offset allow the notification
              to be linked to the alarm that it came from.
             ntype - The notification type
             name - Name used in sending
             address - to send the notification to
             alarm_data - info that caused the notification
             notifications that come after this one to remain uncommitted.
        """
        self.address = address
        self.name = name
        self.src_partition = src_partition
        self.src_offset = src_offset
        self.type = ntype

        self.alarm_id = alarm['alarmId']
        self.alarm_name = alarm['alarmName']
        self.alarm_timestamp = alarm['timestamp']
        self.message = alarm['stateChangeReason']
        self.state = alarm['newState']
        self.tenant_id = alarm['tenantId']
        self.metrics = alarm['metrics']

        # to be updated on actual notification send time
        self.notification_timestamp = None

    def __eq__(self, other):
        if not isinstance(other, Notification):
            return False

        for attrib in self.__slots__:
            if not getattr(self, attrib) == getattr(other, attrib):
                return False

        return True

    def to_json(self):
        """Return json representation
        """
        notification_fields = [
            'address',
            'name',
            'alarm_id',
            'alarm_name',
            'alarm_timestamp',
            'message',
            'notification_timestamp',
            'state',
            'tenant_id'
        ]
        notification_data = {name: getattr(self, name)
                             for name in notification_fields}
        return json.dumps(notification_data)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the notification class."""

import json
from monasca_notification import notification


def test_json():
    """Test the to_json method to verify it behaves as expected.
    """
    alarm = {'alarmId': 'alarmId',
             'alarmName': 'alarmName',
             'timestamp': 'timestamp',
             'stateChangeReason': 'stateChangeReason',
             'newState': 'newState',
             'tenantId': 'tenantId',
             'metrics': 'cpu_util'}
    test_notification = notification.Notification('ntype', 'src_partition', 'src_offset', 'name', 'address', alarm)

    expected_dict = {u'name': u'name',
                     u'notification_timestamp': None,
                     u'tenant_id': u'tenantId',
                     u'alarm_name': u'alarmName',
                     u'alarm_id': u'alarmId',
                     u'state': u'newState',
                     u'alarm_timestamp': u'timestamp',
                     u'address': u'address',
                     u'message': u'stateChangeReason'}
    # Compare as dicts so ordering is not an issue
    assert json.loads(test_notification.to_json()) == expected_dict


def test_equal():
    alarm = {'alarmId': 'alarmId',
             'alarmName': 'alarmName',
             'timestamp': 'timestamp',
             'stateChangeReason': 'stateChangeReason',
             'newState': 'newState',
             'tenantId': 'tenantId',
             'metrics': 'cpu_util'}
    test_notification = notification.Notification('ntype', 'src_partition', 'src_offset', 'name', 'address', alarm)
    test_notification2 = notification.Notification('ntype', 'src_partition', 'src_offset', 'name', 'address', alarm)

    assert(test_notification == test_notification2)


def test_unequal():
    alarm = {'alarmId': 'alarmId',
             'alarmName': 'alarmName',
             'timestamp': 'timestamp',
             'stateChangeReason': 'stateChangeReason',
             'newState': 'newState',
             'tenantId': 'tenantId',
             'metrics': 'cpu_util'}
    test_notification = notification.Notification('ntype', 'src_partition', 'src_offset', 'name', 'address', alarm)
    test_notification2 = notification.Notification('ntype', 'src_partition', 'src_offset', 'name', 'address', alarm)
    test_notification2.alarm_id = None

    assert(test_notification != test_notification2)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import kafka.client
import kafka.producer
import logging
import monascastatsd

from monasca_notification.processors.base import BaseProcessor

log = logging.getLogger(__name__)


class SentNotificationProcessor(BaseProcessor):
    """Processes notifications which have been sent
         This involves adding them into a kafka topic for persisting by another process and adding the alarm
         to the finished queue.
    """

    def __init__(self, sent_queue, finished_queue, url, topic):
        """Init
             url, group - kafka connection details
             topic - kafka topic to publish notifications to
             finished_queue - queue written to when notifications are fully finished.
             sent_queue - the sent_notifications queue notifications are read from
        """
        self.topic = topic
        self.finished_queue = finished_queue
        self.sent_queue = sent_queue
        self.statsd = monascastatsd.Client(name='monasca', dimensions=BaseProcessor.dimensions)

        self.kafka = kafka.client.KafkaClient(url)
        self.producer = kafka.producer.SimpleProducer(
            self.kafka,
            async=False,
            req_acks=kafka.producer.SimpleProducer.ACK_AFTER_LOCAL_WRITE,
            ack_timeout=2000
        )

    def run(self):
        """Takes messages from the sent_queue, puts them on the kafka notification topic and then adds
             partition/offset to the finished queue
        """
        published_to_kafka = self.statsd.get_counter(name='published_to_kafka')

        while True:
            notifications = self.sent_queue.get()
            for notification in notifications:
                responses = self.producer.send_messages(self.topic, notification.to_json())
                published_to_kafka += 1

                log.debug('Published to topic %s, message %s' % (self.topic, notification.to_json()))
                for resp in responses:
                    if resp.error != 0:
                        log.error('Error publishing to %s topic, error message %s' %
                                  (self.topic, resp.error))
            self._add_to_queue(
                self.finished_queue, 'finished', (notifications[0].src_partition, notifications[0].src_offset))



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests NotificationProcessor"""

import json
import mock
import multiprocessing
import requests
import time
import unittest

from monasca_notification.notification import Notification
from monasca_notification.processors import notification_processor


class smtpStub(object):
    def __init__(self, log_queue):
        self.queue = log_queue

    def sendmail(self, from_addr, to_addr, msg):
        self.queue.put("%s %s %s" % (from_addr, to_addr, msg))


class requestsResponse(object):
    def __init__(self, status):
        self.status_code = status


class TestStateTracker(unittest.TestCase):

    def setUp(self):
        self.http_func = self._http_post_200
        self.notification_queue = multiprocessing.Queue(10)
        self.sent_notification_queue = multiprocessing.Queue(10)
        self.finished_queue = multiprocessing.Queue(10)
        self.log_queue = multiprocessing.Queue(10)
        self.email_config = {'server': 'my.smtp.server',
                             'port': 25,
                             'user': None,
                             'password': None,
                             'timeout': 60,
                             'from_addr': 'hpcs.mon@hp.com'}
        self.webhook_config = {'timeout': 50}
        self.pagerduty_config = {'timeout': 50, 'key': 'foobar'}

    def tearDown(self):
        self.assertTrue(self.log_queue.empty())
        self.assertTrue(self.sent_notification_queue.empty())
        self.assertTrue(self.finished_queue.empty())
        self.processor.terminate()

    # ------------------------------------------------------------------------
    # Test helper functions
    # ------------------------------------------------------------------------

    @mock.patch('monasca_notification.processors.notification_processor.requests')
    @mock.patch('monasca_notification.processors.notification_processor.smtplib')
    @mock.patch('monasca_notification.processors.notification_processor.log')
    def _start_processor(self, mock_log, mock_smtp, mock_requests):
        """Start the processor with the proper mocks
        """
        # Since the log runs in another thread I can mock it directly, instead change the methods to put to a queue
        mock_log.warn = self.log_queue.put
        mock_log.error = self.log_queue.put

        mock_requests.post = self.http_func

        mock_smtp.SMTP = self._smtpStub

        self.mock_requests = mock_requests
        self.mock_log = mock_log
        self.mock_smtp = mock_smtp

        nprocessor = (notification_processor.
                      NotificationProcessor(self.notification_queue,
                                            self.sent_notification_queue,
                                            self.finished_queue,
                                            self.email_config,
                                            self.webhook_config,
                                            self.pagerduty_config))

        self.processor = multiprocessing.Process(target=nprocessor.run)
        self.processor.start()

    def _smtpStub(self, *arg, **kwargs):
        return smtpStub(self.log_queue)

    def _http_post_200(self, url, data, headers, **kwargs):
        self.log_queue.put(url)
        self.log_queue.put(data)
        self.log_queue.put(headers)
        r = requestsResponse(200)
        return r

    def _http_post_201(self, url, data, headers, **kwargs):
        self.log_queue.put(url)
        self.log_queue.put(data)
        self.log_queue.put(headers)
        r = requestsResponse(201)
        return r

    def _http_post_202(self, url, data, headers, **kwargs):
        r = requestsResponse(202)
        return r

    def _http_post_204(self, url, data, headers, **kwargs):
        self.log_queue.put(url)
        self.log_queue.put(data)
        self.log_queue.put(headers)
        r = requestsResponse(204)
        return r

    def _http_post_400(self, url, data, headers, **kwargs):
        r = requestsResponse(400)
        return r

    def _http_post_403(self, url, data, headers, **kwargs):
        r = requestsResponse(403)
        return r

    def _http_post_404(self, url, data, headers, **kwargs):
        r = requestsResponse(404)
        return r

    def _http_post_500(self, url, data, headers, **kwargs):
        r = requestsResponse(500)
        return r

    def _http_post_504(self, url, data, headers, **kwargs):
        r = requestsResponse(504)
        return r

    def _http_post_exception(self, url, data, headers, **kwargs):
        self.log_queue.put("timeout %s" % kwargs["timeout"])
        raise requests.exceptions.Timeout

    def assertSentNotification(self):
        notification_msg = self.sent_notification_queue.get(timeout=3)
        self.assertNotEqual(notification_msg, None)

    def assertSentFinished(self):
        finished_msg = self.finished_queue.get(timeout=3)
        self.assertNotEqual(finished_msg, None)

    def alarm(self, metrics):
        return {"tenantId": "0",
                "alarmId": "0",
                "alarmName": "test Alarm",
                "oldState": "OK",
                "newState": "ALARM",
                "stateChangeReason": "I am alarming!",
                "timestamp": time.time(),
                "metrics": metrics}

    def email_setup(self, metric):
        alarm_dict = self.alarm(metric)

        notification = Notification('email', 0, 1, 'email notification', 'me@here.com', alarm_dict)

        self.notification_queue.put([notification])
        self._start_processor()

    def webhook_setup(self, http_func):
        self.http_func = http_func

        metrics = []
        metric_data = {'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        alarm_dict = self.alarm(metrics)

        notification = Notification('webhook', 0, 1, 'email notification', 'me@here.com', alarm_dict)

        self.notification_queue.put([notification])
        self._start_processor()

    def pagerduty_setup(self, http_stub):
        self.http_func = http_stub

        metrics = []
        metric_data = {'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        alarm_dict = self.alarm(metrics)

        notification = Notification('pagerduty',
                                    0,
                                    1,
                                    'pagerduty notification',
                                    'ABCDEF',
                                    alarm_dict)
        self._start_processor()
        self.notification_queue.put([notification])

    def valid_pagerduty_message(self, url, data, headers):
        self.assertEqual(
            url, 'https://events.pagerduty.com/generic/2010-04-15/create_event.json')

        headers = dict(headers)
        self.assertEqual(headers['content-type'], 'application/json')

        data = dict(json.loads(data))
        self.assertEqual(data['service_key'], 'ABCDEF')
        self.assertEqual(data['event_type'], 'trigger')
        self.assertEqual(data['description'], 'I am alarming!')
        self.assertEqual(data['client'], 'Monasca')
        self.assertEqual(data['client_url'], '')

        details = dict(data['details'])
        self.assertEqual(details['alarm_id'], '0')
        self.assertEqual(details['alarm_name'], 'test Alarm')
        self.assertEqual(details['current'], 'ALARM')
        self.assertEqual(details['message'], 'I am alarming!')

    def pagerduty_http_error(self, log_msg, http_response):
        self.assertRegexpMatches(log_msg, "Error with pagerduty request.")
        self.assertRegexpMatches(log_msg, "key=<ABCDEF>")
        self.assertRegexpMatches(log_msg, "response=%s" % http_response)

    # ------------------------------------------------------------------------
    # Unit tests
    # ------------------------------------------------------------------------

    def test_invalid_notification(self):
        """Verify invalid notification type is rejected.
        """
        alarm_dict = {"tenantId": "0", "alarmId": "0", "alarmName": "test Alarm", "oldState": "OK", "newState": "ALARM",
                      "stateChangeReason": "I am alarming!", "timestamp": time.time(), "metrics": "cpu_util"}
        invalid_notification = Notification('invalid', 0, 1, 'test notification', 'me@here.com', alarm_dict)

        self.notification_queue.put([invalid_notification])
        self._start_processor()
        finished = self.finished_queue.get(timeout=2)
        log_msg = self.log_queue.get(timeout=1)
        self.processor.terminate()

        self.assertTrue(finished == (0, 1))
        self.assertTrue(log_msg == 'Notification type invalid is not a valid type')

    def test_email_notification_single_host(self):
        """Email with single host
        """

        metrics = []
        metric_data = {'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)

        self.email_setup(metrics)

        log_msg = self.log_queue.get(timeout=3)

        self.assertRegexpMatches(log_msg, "From: hpcs.mon@hp.com")
        self.assertRegexpMatches(log_msg, "To: me@here.com")
        self.assertRegexpMatches(log_msg, "Content-Type: text/plain")
        self.assertRegexpMatches(log_msg, "Alarm .test Alarm.")
        self.assertRegexpMatches(log_msg, "On host .foo1.")

        self.assertSentNotification()

    def test_email_notification_multiple_hosts(self):
        """Email with multiple hosts
        """

        metrics = []
        metric_data = {'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        self.email_setup(metrics)

        log_msg = self.log_queue.get(timeout=3)

        self.assertRegexpMatches(log_msg, "From: hpcs.mon@hp.com")
        self.assertRegexpMatches(log_msg, "To: me@here.com")
        self.assertRegexpMatches(log_msg, "Content-Type: text/plain")
        self.assertRegexpMatches(log_msg, "Alarm .test Alarm.")
        self.assertNotRegexpMatches(log_msg, "foo1")
        self.assertNotRegexpMatches(log_msg, "foo2")

        self.assertSentNotification()

    def test_webhook_good_http_response(self):
        """webhook 200
        """

        self.webhook_setup(self._http_post_200)

        url = self.log_queue.get(timeout=3)
        data = self.log_queue.get(timeout=3)
        headers = self.log_queue.get(timeout=3)

        self.assertEqual(url, "me@here.com")
        self.assertEqual(data, {'alarm_id': '0'})
        self.assertEqual(headers, {'content-type': 'application/json'})

        self.assertSentNotification()

    def test_webhook_bad_http_response(self):
        """webhook bad response
        """

        self.webhook_setup(self._http_post_404)

        log_msg = self.log_queue.get(timeout=3)

        self.assertNotRegexpMatches(log_msg, "alarm_id.: .test Alarm")
        self.assertNotRegexpMatches(log_msg, "content-type.: .application/json")

        self.assertRegexpMatches(log_msg, "HTTP code 404")
        self.assertRegexpMatches(log_msg, "post on URL me@here.com")

        self.assertSentFinished()

    def test_webhook_timeout_exception_on_http_response(self):
        """webhook timeout exception
        """

        self.webhook_setup(self._http_post_exception)

        log_msg = self.log_queue.get(timeout=3)

        self.assertEqual(log_msg, "timeout 50")

        log_msg = self.log_queue.get(timeout=3)

        self.assertNotRegexpMatches(log_msg, "alarm_id.: .test Alarm")
        self.assertNotRegexpMatches(log_msg, "content-type.: .application/json")

        self.assertRegexpMatches(log_msg, "Error trying to post on URL me@here")
        self.assertRaises(requests.exceptions.Timeout)

        self.assertSentFinished()

    def test_pagerduty_200(self):
        """pagerduty 200
        """

        self.pagerduty_setup(self._http_post_200)

        url = self.log_queue.get(timeout=3)
        data = self.log_queue.get(timeout=3)
        headers = self.log_queue.get(timeout=3)

        self.valid_pagerduty_message(url, data, headers)
        self.assertSentNotification()

    def test_pagerduty_201(self):
        """pagerduty 201
        """

        self.pagerduty_setup(self._http_post_201)

        url = self.log_queue.get(timeout=3)
        data = self.log_queue.get(timeout=3)
        headers = self.log_queue.get(timeout=3)

        self.valid_pagerduty_message(url, data, headers)
        self.assertSentNotification()

    def test_pagerduty_204(self):
        """pagerduty 204
        """

        self.pagerduty_setup(self._http_post_204)

        url = self.log_queue.get(timeout=3)
        data = self.log_queue.get(timeout=3)
        headers = self.log_queue.get(timeout=3)

        self.valid_pagerduty_message(url, data, headers)
        self.assertSentNotification()

    def test_pagerduty_202(self):
        """pagerduty 202
        """

        self.pagerduty_setup(self._http_post_202)
        log_msg = self.log_queue.get(timeout=3)
        self.pagerduty_http_error(log_msg, "202")
        self.assertSentFinished()

    def test_pagerduty_400(self):
        """pagerduty 400
        """

        self.pagerduty_setup(self._http_post_400)
        log_msg = self.log_queue.get(timeout=3)
        self.pagerduty_http_error(log_msg, "400")
        self.assertSentFinished()

    def test_pagerduty_403(self):
        """pagerduty 403
        """

        self.pagerduty_setup(self._http_post_403)
        log_msg = self.log_queue.get(timeout=3)
        self.pagerduty_http_error(log_msg, "403")
        self.assertSentFinished()

    def test_pagerduty_500(self):
        """pagerduty 500
        """

        self.pagerduty_setup(self._http_post_500)
        log_msg = self.log_queue.get(timeout=3)
        self.pagerduty_http_error(log_msg, "500")
        self.assertSentFinished()

    def test_pagerduty_504(self):
        """pagerduty 504
        """

        self.pagerduty_setup(self._http_post_504)
        log_msg = self.log_queue.get(timeout=3)
        self.pagerduty_http_error(log_msg, "504")
        self.assertSentFinished()

    def test_pagerduty_exception(self):
        """pagerduty exception
        """

        self.pagerduty_setup(self._http_post_exception)

        log_msg = self.log_queue.get(timeout=3)

        self.assertEqual(log_msg, "timeout 50")

        log_msg = self.log_queue.get(timeout=3)

        self.assertRegexpMatches(log_msg, "Exception on pagerduty request")
        self.assertRegexpMatches(log_msg, "key=<ABCDEF>")
        self.assertRegexpMatches(
            log_msg, "exception=<class 'requests.exceptions.Timeout'>")

        self.assertRaises(requests.exceptions.Timeout)
        self.assertSentFinished()



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import kafka.client
import kafka.common
import kazoo.client
import kazoo.exceptions
import logging
import Queue
import time

from monasca_notification import notification_exceptions
import monascastatsd

log = logging.getLogger(__name__)


class KafkaStateTracker(object):
    """Tracks message offsets for a kafka topic and partitions.
         As messages are finished with processing the committed offset is updated periodically.
         The messages are not necessarily finished in order, but the committed offset includes
         all previous messages so this object tracks any gaps updating as needed.
         Uses zookeeper to keep track of the last committed offset.
    """
    def __init__(self, finished_queue, kafka_url, group, topic, max_lag, zookeeper_url):
        """Setup the finished_queue
             finished_queue - queue containing all processed alarms
             kafka_url, group, topic - kafka connection details
             zookeeper_url is the zookeeper hostname:port
        """
        self.finished_queue = finished_queue
        self.max_lag = max_lag
        self.topic = topic
        self.has_lock = False
        self.stop = False

        # The kafka client is used solely for committing offsets, the consuming is done in the kafka_consumer
        self.kafka_group = group
        self.kafka = kafka.client.KafkaClient(kafka_url)
        # The consumer is not needed but after my pull request is merged using it for get_offsets simplifies this code
        # import kafka.consumer
        # self.consumer = kafka.consumer.SimpleConsumer(self.kafka, group, topic, auto_commit=False)

        self.zookeeper = kazoo.client.KazooClient(zookeeper_url)
        self.zookeeper.start()

        self.lock_retry_time = 15  # number of seconds to wait for retrying for the lock
        self.lock_path = '/locks/monasca-notification/%s' % topic

        self._offsets = None  # Initialized in the beginning of the run
        self._dimensions = {'service': 'monitoring', 'component': 'monasca-notification'}
        # This is a dictionary of sets used for tracking finished offsets when there is a gap and the committed offset
        # can not yet be advanced
        self._uncommitted_offsets = collections.defaultdict(set)
        self._last_commit_time = collections.defaultdict(time.time)
        statsd = monascastatsd.Client(name='monasca', dimensions=self._dimensions)
        self.offset_update_count = statsd.get_counter(name='alarms_offset_update_count')
        self.finished_count = statsd.get_counter(name='alarms_finished_count')
        self.kafka_timer = statsd.get_timer()

    def _drop_lock(self):
        """Drop the lock file kept in zookeeper
             This should only ever be run when all processing of the finished queue has completed.
        """
        if self.zookeeper.exists(self.lock_path):
            self.zookeeper.delete(self.lock_path)
        self.has_lock = False

    def lock(self, exit_method):
        """Grab a lock within zookeeper, if not available retry.
        """
        while True:
            # The path is ephemeral so if it exists wait then cycle again
            if self.zookeeper.exists(self.lock_path):
                log.info('Another process has the lock for topic %s, waiting then retrying.' % self.topic)
                time.sleep(self.lock_retry_time)
                continue

            try:
                self.zookeeper.create(self.lock_path, ephemeral=True, makepath=True)
            except kazoo.exceptions.KazooException as e:
                # If creating the path fails something beat us to it most likely, try again
                log.warn('Error creating lock path %s\n%s' % (self.lock_path, e))
                continue
            else:
                # Succeeded in grabbing the lock continue
                log.info('Grabbed lock for topic %s' % self.topic)
                self.has_lock = True
                break

        # Set up a listener to exit if we lose connection, this always exits even if the zookeeper connection is only
        # suspended, the process should be supervised so it starts right back up again.
        self.zookeeper.add_listener(exit_method)

    @property
    def offsets(self):
        """Return self._offsets, this is a property because generally only initialize the offsets after the lock has
            been pulled
        """
        if not self.has_lock:
            log.warn('Reading offsets before the tracker has the lock, they could change')

        if self._offsets is None:
            # After my pull request is merged I can setup self.consumer as is done in kafka_consumer
            # then simply use the get_offsets command as below
            # self._offsets = self.consumer.get_offsets()
            self._offsets = {}

            def get_or_init_offset_callback(resp):
                try:
                    kafka.common.check_error(resp)
                    return resp.offset
                except kafka.common.UnknownTopicOrPartitionError:
                    return 0
            for partition in self.kafka.topic_partitions[self.topic]:
                req = kafka.common.OffsetFetchRequest(self.topic, partition)
                (offset,) = self.kafka.send_offset_fetch_request(self.kafka_group, [req],
                                                                 callback=get_or_init_offset_callback,
                                                                 fail_on_error=False)
                self._offsets[partition] = offset

        return self._offsets

    def run(self):
        """Mark a message as finished and where possible commit the new offset to zookeeper.
             There is no mechanism here to deal with the situation where a single alarm is extremely slow to finish
             holding up all others behind it. It is assumed the notification will time out allowing things to finish.
        """
        if not self.has_lock:
            raise notification_exceptions.NotificationException('Attempt to begin run without Zookeeper Lock')

        while True:
            # If self.stop is True run the queue until it is empty, do final commits then exit
            if self.stop and self.finished_queue.empty():
                log.debug('self.stop set and the finished_queue is empty, doing final wait')
                time.sleep(2)  # Before final exit wait a bit to verify the queue is still empty
                if self.finished_queue.empty():
                    if self.max_lag is not None:
                        # if the max_lag has been hit at this point commit the last received offset
                        for partition in self._last_commit_time.iterkeys():
                            if ((time.time() - self._last_commit_time[partition]) > self.max_lag) and\
                                    (len(self._uncommitted_offsets[partition]) > 0):
                                log.error('Max Lag has been reached! Skipping offsets for partition %s' % partition)
                                self.update_offset(partition, max(self._uncommitted_offsets[partition]))
                    break

            try:
                msg = self.finished_queue.get(timeout=10)
            except Queue.Empty:
                continue  # This is non-blocking so the self.stop signal has a chance to take affect

            self.finished_count += 1

            partition = int(msg[0])
            offset = int(msg[1])

            log.debug('Received commit finish for partition %d, offset %d' % (partition, offset))
            # Update immediately if the offset is 1 above current offset
            if self.offsets[partition] == offset - 1:

                new_offset = offset
                for x in range(offset + 1, offset + 1 + len(self._uncommitted_offsets[partition])):
                    if x in self._uncommitted_offsets[partition]:
                        new_offset = x
                        self._uncommitted_offsets[partition].remove(x)
                    else:
                        break

                self.update_offset(partition, new_offset)
                if offset == new_offset:
                    log.debug('Updating offset for partition %d, offset %d' % (partition, new_offset))
                else:
                    log.debug('Updating offset for partition %d, offset %d covering this update and older offsets'
                              % (partition, new_offset))
            elif self.offsets[partition] > offset:
                log.warn('An offset was received that was lower than the committed offset.' +
                         'Possibly a result of skipping lagging notifications')
            else:  # This is skipping offsets so add to the uncommitted set unless max_lag has been hit
                self._uncommitted_offsets[partition].add(offset)
                log.debug('Added partition %d, offset %d to uncommited set' % (partition, offset))
                if (self.max_lag is not None) and ((time.time() - self._last_commit_time[partition]) > self.max_lag):
                    log.error('Max Lag has been reached! Skipping offsets for partition %s' % partition)
                    self.update_offset(partition, max(self._uncommitted_offsets[partition]))
                    self._uncommitted_offsets[partition].clear()

        self._drop_lock()

    def update_offset(self, partition, value):
        """Update the object and kafka offset number for a partition to value
        """
        if self._offsets is None:  # Initialize offsets if needed
            self.offsets

        self.offset_update_count += value - self._offsets[partition]

        self._offsets[partition] = value

        req = kafka.common.OffsetCommitRequest(self.topic, partition, value, None)
        try:
            with self.kafka_timer.time('offset_commit_time_sec'):
                responses = self.kafka.send_offset_commit_request(self.kafka_group, [req])
                kafka.common.check_error(responses[0])

            log.debug('Updated committed offset for partition %s, offset %s' % (partition, value))
        except kafka.common.KafkaError:
            log.exception('Error updating the committed offset in kafka, partition %s, value %s' % (partition, value))
            raise

        self._last_commit_time[partition] = time.time()



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import kafka.client
import kafka.common
import kafka.consumer
import logging
import monascastatsd

from monasca_notification.processors.base import BaseProcessor

log = logging.getLogger(__name__)


class KafkaConsumer(BaseProcessor):
    """Pull from the alarm topic and place alarm objects on the sent_queue.

         No commit is being done until processing is finished and as the processing can take some time it is done in
         another step.

         Unfortunately at this point the python-kafka client does not handle multiple consumers seamlessly.
         For more information see, https://github.com/mumrah/kafka-python/issues/112
    """
    def __init__(self, queue, kafka_url, group, topic):
        """Init
             kafka_url, group, topic - kafka connection details
             sent_queue - a sent_queue to publish log entries to
        """
        self.queue = queue

        self.kafka = kafka.client.KafkaClient(kafka_url)
        # No auto-commit so that commits only happen after the alarm is processed.
        self.consumer = kafka.consumer.SimpleConsumer(self.kafka, group, topic, auto_commit=False)
        self.consumer.provide_partition_info()  # Without this the partition is not provided in the response
        self.statsd = monascastatsd.Client(name='monasca', dimensions=BaseProcessor.dimensions)

        self._initialize_offsets(group, topic)
        # After my pull request is merged I can remove _initialize_offsets and use
        # self.consumer.offsets = self.consumer.get_offsets()
        # self.consumer.fetch_offsets = self.consumer.offsets.copy()
        offsets = self.consumer.offsets.copy()
        self.consumer.seek(0, 0)
        if offsets != self.consumer.offsets:
            log.error('Some messages not yet processed are no longer available in kafka, skipping to first available')
            log.debug('Intialized offsets %s\nStarting offsets %s' % (offsets, self.consumer.offsets))

    def _initialize_offsets(self, group, topic):
        """Fetch initial offsets from kafka
            This is largely taken from what the kafka consumer itself does when auto_commit is used
        """
        def get_or_init_offset_callback(resp):
            try:
                kafka.common.check_error(resp)
                return resp.offset
            except kafka.common.UnknownTopicOrPartitionError:
                return 0

        for partition in self.kafka.topic_partitions[topic]:
            req = kafka.common.OffsetFetchRequest(topic, partition)
            (offset,) = self.consumer.client.send_offset_fetch_request(group, [req],
                                                                       callback=get_or_init_offset_callback,
                                                                       fail_on_error=False)

            # The recorded offset is the last successfully processed, start processing at the next
            # if no processing has been done the offset is 0
            if offset == 0:
                self.consumer.offsets[partition] = offset
            else:
                self.consumer.offsets[partition] = offset + 1

        # fetch_offsets are used by the SimpleConsumer
        self.consumer.fetch_offsets = self.consumer.offsets.copy()

    def run(self):
        """Consume from kafka and place alarm objects on the sent_queue
        """
        consumed_from_kafka = self.statsd.get_counter(name='consumed_from_kafka')

        try:
            for message in self.consumer:
                consumed_from_kafka += 1
                log.debug("Consuming message from kafka, partition %d, offset %d" % (message[0], message[1].offset))
                self._add_to_queue(self.queue, 'alarms', message)
        except Exception:
            log.exception('Error running Kafka Consumer')
            raise
